
def fill_dict_empty_nested(dict_, keys_to_add, nested_key=None):
    """Function to add nested levels in a dictionary.

    Empty nested dictionaries are added for given keywords to be addded.

    Input:
        dict_: dictionary to which nested levels will be added
        nested_key_: (optional) type string, sets the dictionary level
                    to which to add keys (default = None)
       keys_to_add_ = keyword entries to add to dictionary.
    Output:
        None, modifies input dict_
    """
    # TODO: delete?

    if nested_key is None:
        for item in keys_to_add:
            if not dict_ == '':
                dict_[item] = {}
            else:
                dict_ = {item: ""}
    else:
        for item in keys_to_add:
            if not dict_[nested_key] == '':
                dict_[nested_key][item] = {}
            else:
                dict_[nested_key] = {item: ""}




from prettytable import PrettyTable

from h5cross.utils import unwrap_dict


def get_stats_table(stats_dict, full_path=True, digits=6):
    """Function to generate a pretty table output of the added statistics.
    """
    unwrapped_dict = unwrap_dict(stats_dict)  # to avoid nested

    # assumes all have computed same stats
    headers = list(unwrapped_dict[list(unwrapped_dict.keys())[0]]._fields)

    table = PrettyTable(['Dataset'] + headers)

    for key, values in unwrapped_dict.items():
        if not full_path:
            key = key.split('/')[-1]

        table.add_row(_build_row(key, values, headers, digits))

    table.align['Dataset'] = "l"

    return table


def _build_row(key, values, headers, digits=6):
    return [key] + [round(getattr(values, stat), digits) for stat in headers]




import numpy as np
from pymech.dataset import open_dataset

from h5cross.convert._utils import fill_dict_empty_nested


def _add_values_to_dict_from_datasetnek5000(dict_, dataset, skip_list=None):
    """ Function adding nek5000 formatted field variables in a given dictionary
        containing variable names.

        Input:
            :dict_ : empty nested dictionary to fill
            :dataset: nek5000 dataset object
            :skip_list: optional list of type string with dict_ inputs for which
                        to not add the fields
        Output:
            :None: adds fields to the input dict_
    """
    if skip_list is None:
        skip_list = []
    total_len = dataset.dims['x'] * dataset.dims['y'] * dataset.dims['z']

    for key in list(dict_.keys()):
        if key not in skip_list:
            dict_[key] = np.reshape(dataset[key].data, total_len)


def convert_nek5000(file, skip_vars=None):
    """ Main function controlling the conversion of a nek5000 file format
        to dictionary for hdf5 output.

        Input:
            :file: path to nek5000 file
            :skip_vars: optional list of type string containing fields we wish to
                not consider transferring to the hdf5 output
                (default = None)
        Output:
            :new_dict: python dictionary format structured for hdf5 write output
    """

    if skip_vars is None:
        skip_vars = []

    dataset = open_dataset(file)
    var_names = []
    for item in list(dataset.data_vars):
        if item not in skip_vars:
            var_names.append(item)

    new_dict = dict()
    fill_dict_empty_nested(new_dict, var_names)
    _add_values_to_dict_from_datasetnek5000(new_dict, dataset)

    # Add some additional info
    new_dict['Parameters'] = {}
    # print(type(dataset.dims['x']))
    new_dict['Parameters']['Dimensions'] = {'Nx': np.array([dataset.dims['x']], dtype="int32"),
                                            'Ny': np.array([dataset.dims['y']], dtype="int32"),
                                            'Nz': np.array([dataset.dims['z']], dtype="int32")}
    new_dict['Parameters']['time'] = np.array([dataset.coords['time'].data], dtype="float64")

    return new_dict


READER_MAP = {
    'nek5000': convert_nek5000
}




import numpy as np
import vtk

from h5cross.convert._utils import fill_dict_empty_nested


def _add_values_to_dict_from_datasetvtu(dict_, dataset, skip_list=None,
                                        add_mesh=True):
    """ Function adding vtk unstructured formatted field variables in a given dictionary
        containing variable names.

        Input:
            :dict_ : empty nested dictionary to fill
            :dataset: vtk file reader output object
            :skip_list: optional list of type string with dict_ inputs for which
                        to not add the fields
        Output:
            :None: adds fields to the input dict_
    """

    if skip_list is None:
        skip_list = []

    # Field variables part
    pointdata = dataset.GetPointData()
    for variter, varname in enumerate(dict_.keys()):
        if varname not in skip_list:
            array = pointdata.GetArray(variter)
            # Check for multicomponent 1D arrays
            if array.GetNumberOfComponents() > 1:
                # structure in (X0,Y0,Z0) (X1,Y1,Z1) ...
                num_comp = array.GetNumberOfComponents()
                num_values = array.GetNumberOfValues()
                tmp_data = np.zeros((int(num_values / num_comp), num_comp))

                for ii in range(int(num_values / num_comp)):
                    for jj in range(num_comp):
                        tmp_data[ii][jj] = array.GetValue(ii * num_comp + jj)

                comp_names = []
                for jj in range(num_comp):
                    comp_names.append('Component' + str(jj))
                # TODO: review
                fill_dict_empty_nested(dict_, comp_names, nested_key=varname)

                for kk, component in enumerate(comp_names):
                    dict_[varname][component] = np.array(tmp_data[:, kk])

            else:
                tmp_data = []
                for ii in range(array.GetNumberOfValues()):
                    tmp_data.append(array.GetValue(ii))
                dict_[varname] = np.array(tmp_data)

    if add_mesh:
        _add_mesh_values_to_dict_from_datasetvtu(dict_, dataset)


def _add_mesh_values_to_dict_from_datasetvtu(dict_, dataset=None):
    """ Function to add mesh points info from pvtu or vtu file to the given dictionary

        Input:
            :dict_ : empty nested dictionary to fill
            :dataset: vtk file reader output object
        Output:
            :None: adds fields to the input dict_
    """
    # Mesh points
    meshpoints = dataset.GetPoints().GetData()
    num_comp = meshpoints.GetNumberOfComponents()
    num_values = meshpoints.GetNumberOfValues()
    tmp_data = np.zeros((int(num_values / num_comp), num_comp))

    for ii in range(int(num_values / num_comp)):
        for jj in range(num_comp):
            tmp_data[ii][jj] = meshpoints.GetValue(ii * num_comp + jj)
    comp_names = []
    for jj in range(num_comp):
        comp_names.append('Component' + str(jj))

    dict_['Mesh'] = {}
    fill_dict_empty_nested(dict_, comp_names, nested_key='Mesh')
    for kk, component in enumerate(comp_names):
        dict_['Mesh'][component] = np.array(tmp_data[:, kk])


def _get_vtkxml_filetype(filename, readmax_=5):
    """ Function returning the filetype of a vtk xml file.

        Input:
            :filename: input vtk xml file to check
            :readmax_: int, max number of lines of file to search for the type
        Output:
            :myline: str, file type found
            :Exception raised: in case file type not found
    """

    with open(filename, 'r') as fin:
        for _ in range(readmax_):
            myline = fin.readline()
            if "type=" in myline:
                return myline.split("type=")[1].split(' ')[0].split('"')[1]

        raise ValueError("VTK xml File type not found, did you specify the correct file type?")


def convert_vtu(file, parallel=False, skip_vars=None):
    """ Main function controlling the conversion of a vtk xml unstructured file format
        to dictionary for hdf5 output.

        NOTE:   Is currently only based on point data and not cell data.
                No connectivity info is added at the moment.

        Input:
            :file: path to vtk unstructured file
            :skip_vars: optional list of type string containing fields we wish to
                not consider transferring to the hdf5 output
                (default = None)
            :parallel: boolean, specification if we are dealing with pvtu of normal vtu file

        Output:
            :new_dict: python dictionary format structured for hdf5 write output
    """
    # TODO: cell data is ignored for now

    if skip_vars is None:
        skip_vars = []

    # Create reader object
    filetype = _get_vtkxml_filetype(file)
    if parallel:
        assert filetype == 'PUnstructuredGrid', "VTK file type does not coincide with user input"
        reader = vtk.vtkXMLPUnstructuredGridReader()
    else:
        assert filetype == 'UnstructuredGrid', "VTK file type does not coincide with user input"
        reader = vtk.vtkXMLUnstructuredGridReader()

    # Read vtk file
    reader.SetFileName(file)
    reader.Update()

    # Obtain variable names
    vtkdata = reader.GetOutput()
    pointdata = vtkdata.GetPointData()
    vars_num = reader.GetNumberOfPointArrays()
    vars_name = []
    for variter in range(vars_num):
        vars_name.append(pointdata.GetArrayName(variter).split()[0])

    # Create dictionary with structure
    new_dict = dict()
    fill_dict_empty_nested(new_dict, vars_name)

    # Fill dictionary with data
    _add_values_to_dict_from_datasetvtu(new_dict, vtkdata)

    return new_dict


def convert_vtk(file, skip_vars=None):
    """ Main function controlling the conversion of a vtk file format
        to dictionary for hdf5 output.

        NOTE:   Is currently only based on point data and not cell data.
                No connectivity info is added at the moment.
                Currently only for vtk unstructured.

        Input:
            :file: path to vtk unstructured file
            :skip_vars: optional list of type string containing fields we wish to
                not consider transferring to the hdf5 output
                (default = None)

        Output:
            :new_dict: python dictionary format structured for hdf5 write output
    """

    if skip_vars is None:
        skip_vars = []

    # Select reader
    reader = vtk.vtkUnstructuredGridReader()

    # Read vtk file
    reader.SetFileName(file)
    reader.Update()

    # Obtain variable names
    vtkdata = reader.GetOutput()
    pointdata = vtkdata.GetPointData()
    vars_num = pointdata.GetNumberOfArrays()

    vars_name = []
    for variter in range(vars_num):
        vars_name.append(pointdata.GetArrayName(variter).split()[0])

    # Create dictionary with structure
    new_dict = dict()
    fill_dict_empty_nested(new_dict, vars_name)

    # Fill dictionary with data
    _add_values_to_dict_from_datasetvtu(new_dict, vtkdata)

    return new_dict


READER_MAP = {
    'vtk': convert_vtk,
    'vtu': lambda *args, p=False, **kwargs: convert_vtu(*args, parallel=p, **kwargs),
    'pvtu': lambda *args, p=True, **kwargs: convert_vtu(*args, parallel=p, **kwargs)
}



""" Functionalities to convert a certain format to hdf5.
    Currently supported:
        - nek5000 (requires "pymech" package)
        - vtk: vtu, pvtu from ALYA solver, vtk unstructured from OpenFOAM (requires "vtk" package)
"""


from h5cross.convert._convert import convert




import numpy as np
import matplotlib.pyplot as plt
from matplotlib.colors import Normalize
from matplotlib import cm

from h5cross.plotting._utils import _get_density
from h5cross.plotting._utils import _sort_by_density
from h5cross.plotting._utils import _get_scatter_title
from h5cross.plotting._utils import _get_bin_edges


def _plot_scatter(array1, array2, flag_xy_line=False):
    """Function to generate a scatter plot of two data arrays of same length with matplotlib.
    If scipy is available a density scatter plot will be generated.

    Input:
        array1: numpy array of first data set
        array2: numpy array of second data set
        axes_labels: list of string of length 2, used to specify the labels of output axis
        Optional Arguments:
            flag_save_: boolean controlling if the plots get saved (default = False)
            save_name_: (string), base name to be used in saving the plots (default = None)
            flag_show_: boolean setting if the plot gets showed or not (default = True)

    Output:
        Default = None, interactive view
        If flag_save = True then outputs a png image
    """
    # TODO: add histograms

    # Sort the points by density, so that the densest points are plotted last
    density_array = _get_density(array1, array2)
    array1, array2, density_array = _sort_by_density(array1, array2,
                                                     density_array)

    fig, ax = plt.subplots()
    ax.scatter(array1, array2, c=density_array)

    # creates colorbar
    norm = Normalize(vmin=np.min(density_array), vmax=np.max(density_array))
    cbar = fig.colorbar(cm.ScalarMappable(norm=norm), ax=ax)
    cbar.ax.set_ylabel("Density")

    # add line
    if flag_xy_line:
        xmin = np.min(np.append(array1, array2))
        xmax = np.max(np.append(array1, array2))
        ax.plot([xmin, xmax], [xmin, xmax], "k-", linewidth=2)

    textstr = _get_scatter_title(array1, array2)
    ax.set_title(label=textstr, loc="center")

    return ax


def _plot_distributions(array1, array2):
    """Function to generate an histogram of two data arrays of different length with matplotlib

    Input:
        array1: numpy array of first data set
        array2: numpy array of second data set
        Optional Arguments:
            flag_save_: boolean controlling if the plots get saved (default = False)
            save_name_: (string), base name to be used in saving the plots (default = None)
            flag_show_: boolean setting if the plot gets showed or not (default = True)

    Output:
        Default = None, interactive view
        If flag_save = True then outputs a png image
    """

    _, ax = plt.subplots()

    bin_edges = _get_bin_edges(array1, array2)

    for i, (array, color, alpha) in enumerate(zip(
            [array1, array2], ["cornflowerblue", "lightgreen"], [0.8, 0.6])):
        ax.hist(array, bins=bin_edges, density=False, facecolor=color, alpha=alpha,
                label=f"field {i}")

    ax.legend()

    return ax



"""Module with tools to recursively compare nested objects
BEWARE of stack overflows in recursions"""

import numpy as np
import h5py

from h5cross import hdfdict


DATA_TYPES = (
    np.short,
    np.ushort,
    np.int8,
    np.uint8,
    np.intc,
    np.uintc,
    np.int16,
    np.uint16,
    np.uint,
    np.int_,
    np.int32,
    np.uint32,
    np.longlong,
    np.ulonglong,
    np.uint64,
    np.int64,
    np.float16,
    np.half,
    np.float32,
    np.single,
    np.float64,
    np.double,
    np.longdouble,
    np.float_,
    np.complex_,
    float,
    int,
    complex
)

DESC_TYPES = (np.bytes_, np.string_, str, bytes, np.bool_, bool)


def h5dict_safe(file: str) -> dict:
    """
    h5 files safe loader with hdfdict.
    :param file: str ou os.Path menant au h5 à lire
    """
    with h5py.File(file, 'r') as fin:
        try:
            return hdfdict.load(fin, lazy=False)
        except TypeError:
            return hdfdict.load(fin)


def h5_same(source_file: str, target_file: str, **kwargs):
    """
    *Main call function to test two h5 files.*

    :param source: Path to the source file to compare
    :param target: Path to the target file to compare
    :returns: **True** if files are identical, **False** otherwise
    """
    source_dict = h5dict_safe(source_file)
    target_dict = h5dict_safe(target_file)
    try:
        return dict_same(source_dict, target_dict, **kwargs)
    except ValueError as err:
        raise ValueError(
            f"\nSource file is {source_file}\nTarget file is {target_file}\n" + str(err)
        )


def dict_same(source_dict: dict, target_dict: dict, **kwargs):
    """
    *Main call function to test two dictionaries.*

    :param source_dict: first dict to compare
    :param target_dict: second dict to compare
    :returns: **True** if files are identical, raises an error otherwise
    """
    is_identical, log = _same_nob(source_dict, target_dict, **kwargs)
    if not is_identical:
        raise ValueError(log)
    return True


def _same_nob(src: dict, tgt: dict, path="/") -> bool:
    """Compare NOb identified as same, numpy arrays included, recursively"""
    # pylint: disable=too-many-locals,too-many-branches,too-many-statements
    log = str()
    log += "------------------------------\n"
    log += "Examining " + path + "\n"
    output = True

    common = dict()

    # compare keys
    for key in src:
        if not key in tgt:
            output = False
            log += f"Attribute {key} only in source h5 file.\n"
        else:
            common[key] = src[key]
    for key in tgt:
        if not key in src:
            output = False
            log += f"Attribute {key} only in target h5 file.\n"

    # compare whats inside
    for key in common:
        log += "\t" + key + "\n"
        tgt_t = type(tgt[key])
        src_t = type(src[key])
        if not src_t == tgt_t:
            output = False
            log += (
                f"**  Different attribute types: {src_t} and {tgt_t} (DIFF_OBJECTS)\n"
            )
            continue  # different hdf5 types -- don't try to compare further

        # handle datasets first
        if isinstance(tgt[key], (DESC_TYPES, DATA_TYPES)):
            if not src[key] == tgt[key]:
                output = False
                log += f"** Different values for elements {key}: {src[key]} and {tgt[key]}\n"

        elif isinstance(tgt[key], (list, tuple, np.ndarray)):
            if np.shape(tgt[key]) != np.shape(src[key]):
                output = False
                log += f" ** Different shapes for arrays {key}"
            elif isinstance(tgt[key][0], (np.ndarray, DATA_TYPES)):
                if not np.allclose(src[key], tgt[key]):
                    output = False
                    log += f"** Different values in arrays {key}\n"
            elif isinstance(tgt[key][0], DESC_TYPES):
                if not np.all(src[key] == tgt[key]):
                    output = False
                    log += f"** Different values in arrays {key}: {src[key]} and {tgt[key]}\n"
            else:
                raise NotImplementedError(
                    f"Attribute's values type: {type(tgt[key][0])}"
                )

        # handle nob recursively
        elif isinstance(tgt[key], dict):
            subout, sublog = _same_nob(src[key], tgt[key], path=path + key + "/")
            log += sublog
            if subout is False:
                output = subout
        else:
            raise NotImplementedError(f"Attribute type: {type(tgt[key])}")

    return output, log





import matplotlib.pyplot as plt
from nob import Nob

from h5cross.io import load_hdf_as_dict
from h5cross.plotting._utils import _save_plot
from h5cross.plotting._utils import _load_backend

DEFAULT_BACKEND = 'matplotlib'


# TODO: rotate axis labels in right histogram


def compare_scatter_h5(file1, file2, variable_list, save_name=None,
                       flag_show=True, flag_seaborn=True, flag_xy_line=False):
    """Function that controls the generation of scatterplots comparing the same
    variable from two hdf5 files

    Input:
        :file1: (string), name of first file for comparison
        :file2: (string), name of second file for comparison
        :variable_list: list of type string containing keywords of the
                        variables to select for comparison. The same variable
                        will be selected from each file.
        Optional Arguments:
            :flag_save: boolean controlling if the plots get saved (default = False)
            :save_name: (string), base name to be used in saving the plots (default = None)
            :flag_show: boolean setting if the plot gets showed or not (default = True)
            :flag_seaborn: boolean controlling whether seaborn is used or not (default = True)
    Output:
        :None: interactive view. If flag_save = True, a png image of the scatter plots gets saved.
    """
    # TODO: keep flag?
    backend = 'seaborn' if flag_seaborn else 'matplotlib'

    nob1 = Nob(load_hdf_as_dict(file1))
    nob2 = Nob(load_hdf_as_dict(file2))
    for var_name in variable_list:
        array1 = nob1[var_name][:]
        array2 = nob2[var_name][:]
        if (array1 is not None) and (array2 is not None):
            if len(array1) == len(array2):
                plot_scatter(array1, array2, axes_labels=[var_name, var_name],
                             flag_show=False, save_name=save_name,
                             flag_xy_line=flag_xy_line, backend=backend)
            else:
                print("Warning: data arrays have different length, plotting distribution instead")

                plot_distributions(array1, array2, axes_labels=[var_name, var_name],
                                   flag_show=False,
                                   save_name=save_name, backend=backend)

    if flag_show:
        plt.show()


def plot_scatter(*args, axes_labels=None, save_name=None, flag_show=True,
                 backend=DEFAULT_BACKEND, **kwargs):
    module = _load_backend(backend)

    ax = module._plot_scatter(*args, **kwargs)

    if axes_labels is not None:
        x_label = f'{axes_labels[0]} field 1'
        y_label = f'{axes_labels[1]} field 2'
    else:
        x_label = 'field 1'
        y_label = 'field 2'
    ax.set_xlabel(x_label)
    ax.set_ylabel(y_label)
    plt.tight_layout()

    if save_name:
        _save_plot(save_name, axes_labels)

    if flag_show:
        plt.show()

    return ax


def plot_distributions(*args, axes_labels=None, save_name=None, flag_show=True,
                       backend=DEFAULT_BACKEND, **kwargs):
    module = _load_backend(backend)

    ax = module._plot_distributions(*args, **kwargs)

    x_label = 'field 1' if axes_labels is None else f'{axes_labels[0]} field 1'
    ax.set_xlabel(x_label)
    ax.set_ylabel("Count")
    plt.tight_layout()

    if save_name:
        _save_plot(save_name, axes_labels)

    if flag_show:
        plt.show()

    return ax



"""Module with functionalities related to the visual aspects of h5cross.
"""

from nobvisual.nob2nstruct import visual_treenob
from nobvisual.nobcompare import nob_compare_tkinter

from h5cross.visit_h5 import get_h5_structure
from h5cross.io import load_hdf_as_dict
from h5cross.stats import compute_dict_stats
from h5cross.utils import merge_dicts


def compare_h5(file1, file2, add_stats=False, start_mainloop=True):
    """Function allowing the comparison of two hdf5 files

    Input:
        :file1: (string), name of first file for comparison
        :file2: (string), name of second file for comparison
        :add_stats: (boolean), option to compute statistics on both input files
    Output:
        :None: interactive comparative view
    """

    # TODO: write separate functionalities for this?? (Jimmy)
    nobs = []
    for file in [file1, file2]:
        dict_ = get_h5_structure(file, as_dict=not add_stats)

        if add_stats:
            data_dict = load_hdf_as_dict(file)
            stats = compute_dict_stats(data_dict)
            dict_ = merge_dicts(dict_, stats, as_dict=True)

        nobs.append(dict_)

    title = "Showing file differences"
    title += f"\nLeft: {file1}"
    title += f"\nRight: {file2}"
    nob_compare_tkinter(*nobs, title=title, start_mainloop=start_mainloop)


def visual_h5(filename, start_mainloop=True):
    """Function calling nobvisual for visual view of hdf5 file

    Input:
        :filename: name of hdf5 file
    Output:
        :None: interactive view
    """

    nob = get_h5_structure(filename)
    visual_treenob(nob, title=f'Showing {filename}',
                   start_mainloop=start_mainloop)



"""
Python tool to test if h5 files are identical.
"""
import h5py
import filecmp
import difflib
import pytest
import os

# TODO: improve logging
# TODO: shouldn't this be based on hdfdict?


def _read_attributes(hval):
    """Return the attribute."""
    attr = {}
    for k in hval.attrs:
        attr[k] = type(hval.attrs[k])
    return attr


def _read_group(hval):
    """ returns summary of group.
    the only element for comparison here is the group's attributes
    """
    desc = {}
    desc["attr"] = _read_attributes(hval)
    desc["htype"] = "group"
    return desc


def _read_data(hval):
    """ returns summary of dataset
    the only elements for comparison here are the dataset's attributes,
    and the dataset dtype"""
    desc = {}
    desc["attr"] = _read_attributes(hval)
    desc["htype"] = "dataset"
    desc["dtype"] = type(hval[()])
    return desc


def _evaluate_group(path, grp):
    """creates and returns a summary description
    for every element in a group
    """
    desc = {}
    for key, val in grp.items():
        if isinstance(val, h5py.Dataset):
            desc[key] = _read_data(val)
        elif isinstance(val, h5py.Group):
            desc[key] = _read_group(val)
        else:
            raise NotImplementedError(
                "Unknown h5py type: %s (%s -- %s)" %
                (type(val), path, key))
    return desc


def _same_groups(file1, grp1, file2, grp2, path):
    """Compare groups idientified as same."""
    log = str()
    log += ("------------------------------\n")
    log += ("Examining " + path + "\n")
    output = True

    desc1 = _evaluate_group(path, grp1)
    desc2 = _evaluate_group(path, grp2)
    common = []
    for k in desc1:
        if k in desc2:
            common.append(k)
        else:
            output = False
            log += ("** Element '%s' only in '%s' (DIFF_UNIQUE_A)**\n" %
                    (k, file1))
    for k in desc2:
        if k not in desc1:
            output = False
            log += ("** Element '%s' only in '%s' (DIFF_UNIQUE_B)**\n" %
                    (k, file2))
    for i, _ in enumerate(common):
        name = common[i]
        log += ("\t" + name + "\n")
        # compare types
        htype_1 = desc1[name]["htype"]
        htype_2 = desc2[name]["htype"]
        if htype_1 != htype_2:
            output = False
            log += (
                "**  Different element types: "
                + "'%s' and '%s' (DIFF_OBJECTS)\n" % (htype_1, htype_2))
            continue    # different hdf5 types -- don't try to compare further
        if htype_1 not in ("dataset", "group"):
            log += (
                "WARNING: element is not a recognized type"
                + " (%s) and isn't being evaluated\n" % htype_1)
            continue
        # handle datasets first
        if desc1[name]["htype"] != "dataset":
            continue
        # compare data
        if desc1[name]["dtype"] != desc2[name]["dtype"]:
            dtype_1 = desc1[name]["dtype"]
            dtype_2 = desc2[name]["dtype"]
            output = False
            log += ("** Different dtypes: '%s' and '%s' (DIFF_DTYPE)**\n" % (dtype_1, dtype_2))
        # compare attributes
        for k in desc1[name]["attr"]:
            if k not in desc2[name]["attr"]:
                output = False
                log += ("** Attribute '%s' only in '%s' (DIFF_UNIQ_ATTR_A)**\n" %
                        (k, file1))
        for k in desc2[name]["attr"]:
            if k not in desc1[name]["attr"]:
                output = False
                log += ("** Attribute '%s' only in '%s' (DIFF_UNIQ_ATTR_B)**\n" %
                        (k, file2))
        for k in desc1[name]["attr"]:
            if k in desc2[name]["attr"]:
                val = desc1[name]["attr"][k]
                val2 = desc2[name]["attr"][k]
                if val != val2:
                    output = False
                    log += (
                        "** Attribute '%s' has different type: '%s' and '%s' (DIFF_ATTR_DTYPE)\n" %
                        (k, val, val2))

    for i, _ in enumerate(common):
        name = common[i]
        # compare types
        if desc1[name]["htype"] != desc2[name]["htype"]:
            continue    # problem already reported
        if desc1[name]["htype"] != "group":
            continue
        # compare attributes
        for k in desc1[name]["attr"]:
            if k not in desc2[name]["attr"]:
                output = False
                log += ("** Attribute '%s' only in '%s' (DIFF_UNIQ_ATTR_A)**\n" %
                        (k, file1))
        for k in desc2[name]["attr"]:
            if k not in desc1[name]["attr"]:
                output = False
                log += ("** Attribute '%s' only in '%s' (DIFF_UNIQ_ATTR_B)**\n" %
                        (k, file2))
        # recurse into subgroup
        sub_log, sub_out = _same_groups(
            file1, grp1[name], file2, grp2[name], path + name + "/")
        log += sub_log
        if sub_out is False:
            output = False

    return log, output


def h5same_files(file1: str, file2: str):
    """
    *Main call function to test two h5 files.*

    :param file1: Path of the first file to compare
    :type file1: str
    :param file2: Path of the second file to compare
    :type file2: str
    :returns: **True** if files are identical, **False** otherwise
    """
    fin_1 = h5py.File(file1, 'r')
    fin_2 = h5py.File(file2, 'r')
    return _same_groups(file1, fin_1["/"], file2, fin_2["/"], "/")



def compare_textfile(txtfile1: str, txtfile2: str):
    """Small utility to better format diff lib"""
    try:
        assert filecmp.cmp(txtfile1, txtfile2)
    except AssertionError:
        with open(txtfile1, "r") as fin:
            str1 =fin.readlines()
        with open(txtfile2, "r") as fin:
            str2 = fin.readlines()
        diff_ = list()
        for line in difflib.ndiff(str1, str2):
            for mark in ["+ ", "- ", "? "]:
                if line.startswith(mark):
                    diff_.append(line)
                    print(">>>>", line)
        pytest.fail(
            "Comparing text files :\n"
            + "..."+ os.path.split(txtfile1)[1] + "\n"
            + "..."+ os.path.split(txtfile2)[1] + "\n"
            + '\n'.join(diff_))




import importlib

import numpy as np
from scipy.interpolate import interpn
import matplotlib.pyplot as plt


def _load_backend(backend):
    return importlib.import_module(f'h5cross.plotting._{backend}')


def _get_density(x_array, y_array, bins=24):
    data, xedge, yedge = np.histogram2d(x_array, y_array, bins=bins, density=True)

    density_array = interpn(
        (0.5 * (xedge[1:] + xedge[:-1]), 0.5 * (yedge[1:] + yedge[:-1])),
        data, np.vstack([x_array, y_array]).T, method="splinef2d", bounds_error=False)

    # To be sure to plot all data
    # TODO: review? use case? probably undesired?
    density_array[np.where(np.isnan(density_array))] = 0.0

    return density_array


def _sort_by_density(x_array, y_array, density_array):
    idx = density_array.argsort()
    return x_array[idx], y_array[idx], density_array[idx]


def _get_error_norm(array1, array2, order=2):
    error_norm = np.linalg.norm(array1 - array2, ord=order)
    error_norm /= np.linalg.norm((array1 + array2) / 2.0, ord=order)

    return error_norm


def _get_scatter_title(array1, array2):
    error_norm = _get_error_norm(array1, array2, order=2)
    return f"Similarity: {1-error_norm:.4f}"


def _get_bin_edges(array1, array2):
    bin_edges, _ = np.lib.histograms._get_bin_edges(np.r_[array1, array2], 'auto',
                                                    None, None)

    return bin_edges


def _save_plot(filename, axes_labels=None):
    if axes_labels:
        filename = f'{filename}_{axes_labels[0]}_{axes_labels[1]}'

    plt.savefig(f'{filename}.png')



"""Module controlling YAML and hdf5 output files.
"""


import yaml
import h5py

from h5cross import hdfdict


def load_hdf_as_dict(h5_filename, lazy=False):
    """Return a dictionary of an hdf5 file.
    """
    with h5py.File(h5_filename, "r") as h5pf:
        data_dict = hdfdict.load(h5pf, lazy=False)
    return data_dict


def write_dict_to_yaml(dict_, filename, setindent=2):
    """ Generic function writin a dictionary as a yaml file

        Input:
            :dict_: dictionary to write out as yaml
            :filename: output name of the .yml file
            :setindent: indentation setting of yaml file
        Output:
            a YAML file
    """
    with open(f'{filename}.yml', 'w') as yaml_out:
        yaml.dump(dict_, yaml_out, indent=setindent)


def write_h5(dict_, filename='dump'):
    with h5py.File(f'{filename}.h5', "w") as fout:
        hdfdict.dump(dict_, fout, lazy=False)



""" Command line module of h5cross"""

import click
import h5cross


_yaml_filename_option = click.option(
    "--output-filename", "-o", type=str, default=None,
    help='Yaml output filename.')


def add_options(options):
    # https://stackoverflow.com/a/40195800/11011913
    def _add_options(func):
        for option in reversed(options):
            func = option(func)
        return func
    return _add_options

def add_version(f):
    """
    Add the version of the tool to the help heading.
    :param f: function to decorate
    :return: decorated function
    """
    doc = f.__doc__
    f.__doc__ = "Package " + h5cross.__name__ + " v" + h5cross.__version__ + "\n\n" + doc

    return f

@click.group()
@add_version
def main_cli():
    """---------------    H5CROSS  --------------------

You are now using the Command line interface of h5cross
a Python3 helper to explore and compare hdf5 files, created at CERFACS (https://cerfacs.fr).


This is a python package currently installed in your python environment.
"""


@click.command()
@click.argument("filename", nargs=1)
def tree(filename):
    """Print the content of an hdf5 file in terminal.
    """
    from h5cross.visit_h5 import get_h5_structure
    from h5cross.utils import print_dict

    print_dict(get_h5_structure(filename))


main_cli.add_command(tree)


@click.command()
@click.argument("filename", nargs=1)
@add_options([_yaml_filename_option])
def dump(filename, output_filename):
    """Write the content of an hdf5 file into a YAML file.

    """
    from h5cross.visit_h5 import get_h5_structure
    from h5cross.io import write_dict_to_yaml

    mydict = get_h5_structure(filename)
    writename = filename.split('.')[0] if output_filename is None else output_filename

    write_dict_to_yaml(mydict, writename)


main_cli.add_command(dump)


@click.command()
@click.argument("filename", nargs=1)
@click.option('--test', is_flag=True, hidden=True)
def view(filename, test):
    """Show the content of an hdf5 file interactively with nobvisual.
    """
    from h5cross.nobvisual import visual_h5

    visual_h5(filename, start_mainloop=not test)


main_cli.add_command(view)


@click.command()
@click.argument("filename", nargs=1)
@click.option('--pretty', is_flag=True,
              help='Controls if pretty table output is used')
@click.option('--pretty-full-path', is_flag=True,
              help='Controls if full paths are shown in pretty table')
@click.option('--save-as-yaml', is_flag=True,
              help='Default output name is set to input file name')
@add_options([_yaml_filename_option])
def stats(filename, pretty, pretty_full_path, save_as_yaml, output_filename):
    """Compute statistics of arrays from hdf5 file.
    """
    from h5cross.visit_h5 import get_h5_structure
    from h5cross.stats import compute_dict_stats
    from h5cross.io import load_hdf_as_dict
    from h5cross.io import write_dict_to_yaml
    from h5cross.utils import merge_dicts
    from h5cross.utils import print_dict

    data_dict = load_hdf_as_dict(filename)
    stats = compute_dict_stats(data_dict)

    if save_as_yaml or (not pretty):
        h5_struct = get_h5_structure(filename, as_dict=False)
        new_dict = merge_dicts(h5_struct, stats, as_dict=True)

    if pretty:
        from h5cross.prettytable import get_stats_table
        stats_table = get_stats_table(stats, full_path=pretty_full_path)
        print(stats_table)

    else:
        print_dict(new_dict)

    if save_as_yaml:
        writename = filename.split('.')[0] if output_filename is None else output_filename
        write_dict_to_yaml(new_dict, writename)


main_cli.add_command(stats)


@click.command()
@click.argument("file_left", nargs=1)
@click.argument("file_right", nargs=1)
@click.option('--add-stats', is_flag=True,
              help='It computes statistics for each file before comparing')
@click.option('--test', is_flag=True, hidden=True)
def diff(file_left, file_right, add_stats, test):
    """Compare the content of two hdf5 files and view interactively with nobvisual.
    """
    from h5cross.nobvisual import compare_h5

    compare_h5(file_left, file_right, add_stats, start_mainloop=not test)


main_cli.add_command(diff)


@click.command()
@click.argument("file_left", nargs=1)
@click.argument("file_right", nargs=1)
@click.option('--select-vars', required=True, default=None, type=str,
              help='default = None, select variables to plot. Requires a comma separated string with \
            variables to select, e.g. \"temperature,pressure\" ')
@click.option('--save-name', required=False, default=None, type=str,
              help='default = None, string specifying desired output base name for each plot')
@click.option('--use-seaborn', is_flag=True,
              help='Controls the use of seaborn, otherwise matplotlib is used.')
@click.option('--add-xy', is_flag=True,
              help='Superimposes the x=y line on the scatter plot.')
@click.option('--test', is_flag=True, hidden=True)
def scatter(file_left, file_right, select_vars, save_name, use_seaborn, add_xy,
            test):
    """ Scatter plot comparison of two hdf5 files.
        Seaborn is used to generate the plots but can
        be deactivated in which case matplotlib.pyplot is used.

        Note: The matplotlib package is a minimal requirement for this functionality.

    """
    from h5cross.plotting import compare_scatter_h5

    var_list = select_vars.split(",")
    compare_scatter_h5(file_left, file_right, var_list, save_name=save_name,
                       flag_show=not test, flag_seaborn=use_seaborn,
                       flag_xy_line=add_xy)


main_cli.add_command(scatter)


@click.command()
@click.argument("filename", nargs=1)
@click.option('--file-type', required=True, type=click.Choice(['nek5000', 'pvtu', 'vtu', 'vtk']), default=None,
              help='default = None, input format to convert to hdf5')
@click.option('--out-name-h5', required=False, default=None, type=str,
              help='Optional output name for hdf5 file, default = dump.h5')
@click.option('--out-location-h5', required=False, default=None, type=str,
              help='Optional path where to save the hdf5 file, default = current dir')
def convert(filename, file_type, out_name_h5, out_location_h5):
    """ Conversion to hdf5 of certain file formats.
        Currently supported:
            - nek5000: requires "pymech" package
            - vtk: pvtu, vtu   requires "vtk" package
    """
    # TODO: write_h5
    # TODO: change flags and function: do not distinguish between path and filename

    from h5cross.convert import convert
    from h5cross.io import write_h5

    tmp_dict = convert(filename, file_type)

    if out_name_h5 is not None:
        if out_location_h5 is not None:
            write_h5(tmp_dict, save_name_=out_name_h5, save_path_=out_location_h5)
        else:
            write_h5(tmp_dict, save_name_=out_name_h5)
    else:
        if out_location_h5 is not None:
            write_h5(tmp_dict, save_path_=out_location_h5)
        else:
            write_h5(tmp_dict)


main_cli.add_command(convert)




from collections import namedtuple

import numpy as np

from h5cross.utils import unwrap_dict
from h5cross.utils import wrap_dict


STATS = {
    'min': np.min,
    'max': np.max,
    'mean': np.mean,
    'median': np.median,
    'std': np.std
}


def compute_stat(array, stat):
    return float(STATS[stat](array))


def compute_stats(array, stats):
    StatsCollection = namedtuple('StatsCollection', stats)
    return StatsCollection(*[compute_stat(array, stat) for stat in stats])


def compute_dict_stats(dict_, skip_list=(),
                       stats_list=('min', 'max', 'mean', 'median', 'std')):

    unwrapped_dict = unwrap_dict(dict_)  # to avoid nested

    stats = {}
    for key, array in unwrapped_dict.items():
        if key.split('/')[-1] in skip_list:
            continue

        try:
            stats[key] = compute_stats(array, stats_list)
        except TypeError:
            pass

    # wrap again
    return wrap_dict(stats)




from collections import namedtuple

import yaml


def print_dict(dict_):
    print(yaml.dump(dict_, default_flow_style=False))


def unwrap_dict(dict_, sep='/'):
    new_dict = {}
    _unwrap_item(dict_, '', new_dict, sep=sep)

    return new_dict


def wrap_dict(unwrapped_dict, sep='/', operate_on_value=None):
    new_keys = _group_keys(unwrapped_dict.keys(), sep=sep)
    return _wrap_item(unwrapped_dict, new_keys, sep=sep,
                      operate_on_value=operate_on_value)


def _unwrap_item(dict_, parent_key, new_dict, sep='/'):

    for key, value in dict_.items():
        if parent_key:
            new_key = f'{parent_key}{sep}{key}'
        else:
            new_key = key

        if type(value) is dict:
            _unwrap_item(value, new_key, new_dict,
                         sep=sep)
        else:
            new_dict[new_key] = value


def _wrap_item(unwrapped_dict, new_keys, sep='/', operate_on_value=None):
    new_dict = {}

    for key, key_value in new_keys.items():
        if type(key_value) is dict:
            value = _wrap_item(unwrapped_dict, key_value, sep=sep,
                               operate_on_value=operate_on_value)
        else:
            value = unwrapped_dict[key_value]

            if callable(operate_on_value):
                value = operate_on_value(value)

        new_dict[key] = value

    return new_dict


def _group_keys_one_level(group, sep='/'):
    groups = {}

    for key, value in group.items():
        key_split = key.split(sep)

        if len(key_split) == 1:
            groups[key] = value
            continue

        start_key = f'{sep}'.join(key_split[:1])
        end_key = f'{sep}'.join(key_split[1:])

        if start_key not in groups:
            groups[start_key] = {}

        groups[start_key][end_key] = value

    return groups


def _group_keys_recursively(group, sep='/'):
    nested_group = _group_keys_one_level(group, sep=sep)

    new_group = {}
    for key, group_ in nested_group.items():
        if type(group_) is dict:
            new_group[key] = _group_keys_recursively(group_, sep=sep)
        else:
            new_group[key] = group_

    return new_group


def _group_keys(keys, sep='/'):

    group = {key: key for key in keys}
    groups = _group_keys_recursively(group, sep=sep)

    return groups


def _merge_item(dict_, stats):
    # TODO: pass as_dict
    # TODO: if not stats, no merging
    new_dict = {}

    for key, value in dict_.items():
        if type(value) is dict:
            value_ = _merge_item(value, stats)
        else:

            if isinstance(value, tuple):
                value_ = value._asdict()

        new_dict[key] = value_

    return new_dict


def merge_dicts(dict_left, dict_right, name_left='left_value',
                name_right='right_value', as_dict=False):
    """Merges structure dict with stats dict.

    Assumes last element is namedtuple or value. If namedtuple, `name_left` or
    `name_right` are ignored.

    If `as_dict`, then merge is `dict` instead of `namedtuple`.
    """
    unwrapped_left = unwrap_dict(dict_left)
    unwrapped_right = unwrap_dict(dict_right)

    new_dict = {}

    # act on left
    for key, left_value in unwrapped_left.items():

        if isinstance(left_value, tuple):
            value = left_value._asdict()
        else:
            value = {name_left: left_value}

        right_value = unwrapped_right.get(key, None)
        if right_value is not None:
            if isinstance(right_value, tuple):
                value.update(right_value._asdict())
            else:
                value.update({name_right: right_value})

        new_dict[key] = get_namedtuple_from_dict(value)

    # act on right
    for key, right_value in unwrapped_right.items():
        if key in unwrapped_left:
            continue

        if isinstance(right_value, tuple):
            value = right_value._asdict()
        else:
            value = {name_right: right_value}

        new_dict[key] = get_namedtuple_from_dict(value)

    # wrap again
    operate_on_value = get_dict_from_namedtuple if as_dict else None
    return wrap_dict(new_dict, operate_on_value=operate_on_value)


def get_namedtuple_from_dict(dict_, name='Container'):
    NamedTuple = namedtuple(name, dict_)
    return NamedTuple(**dict_)


def get_dict_from_namedtuple(namedtuple_):
    return namedtuple_._asdict()



"""Helps h5py to dump and load dictionaries.

Adapted from https://github.com/SiggiGue/hdfdict due to inactivity of the
developer.
"""


from collections import UserDict
from datetime import datetime
from contextlib import contextmanager

import h5py
import yaml
from numpy import string_


TYPEID = '_type_'


@contextmanager
def _hdf_file(hdf, lazy=True, *args, **kwargs):
    """Context manager yields h5 file if hdf is str,
    otherwise just yield hdf as is."""
    if isinstance(hdf, str):
        if not lazy:
            with h5py.File(hdf, *args, **kwargs) as hdf:
                yield hdf
        else:
            yield h5py.File(hdf, *args, **kwargs)
    else:
        yield hdf


def _unpack_dataset(item):
    """Reconstruct a hdfdict dataset.
    Only some special unpacking for yaml and datetime types.

    Parameters
    ----------
    item : h5py.Dataset

    Returns
    -------
    value : Unpacked Data

    """
    value = item[()]

    if TYPEID in item.attrs:
        if item.attrs[TYPEID].astype(str) == 'datetime':
            if hasattr(value, '__iter__'):
                value = [datetime.fromtimestamp(
                    ts) for ts in value]
            else:
                value = datetime.fromtimestamp(value)

        if item.attrs[TYPEID].astype(str) == 'yaml':
            value = yaml.safe_load(value.decode())

    return value


def _pack_dataset(hdfobject, key, value):
    """Packs a given key value pair into a dataset in the given hdfobject."""
    isdt = None
    if isinstance(value, datetime):
        value = value.timestamp()
        isdt = True

    if hasattr(value, '__iter__'):
        if all(isinstance(i, datetime) for i in value):
            value = [item.timestamp() for item in value]
            isdt = True

    try:
        ds = hdfobject.create_dataset(name=key, data=value)
        if isdt:
            ds.attrs.create(
                name=TYPEID,
                data=string_("datetime"))
    except TypeError:
        # Obviously the data was not serializable. To give it
        # a last try; serialize it to yaml
        # and save it to the hdf file:
        ds = hdfobject.create_dataset(
            name=key,
            data=string_(yaml.safe_dump(value))
        )
        ds.attrs.create(
            name=TYPEID,
            data=string_("yaml"))
        # if this fails again, restructure your data!


class LazyHdfDict(UserDict):
    """Helps loading data only if values from the dict are requested.

    This is done by reimplementing the __getitem__ method.
    """

    def __init__(self, _h5file=None, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._h5file = _h5file  # used to close the file on deletion.

    def __getitem__(self, key):
        """Returns item and loads dataset if needed."""
        item = super().__getitem__(key)
        if isinstance(item, h5py.Dataset):
            item = _unpack_dataset(item)
            self.__setitem__(key, item)
        return item

    def unlazy(self):
        """Unpacks all datasets.
        You can call dict(this_instance) then to get a real dict.
        """
        load(self, lazy=False)

    def close(self):
        """Closes the h5file if provided at initialization."""
        if self._h5file and hasattr(self._h5file, 'close'):
            self._h5file.close()

    def __del__(self):
        self.close()

    def _ipython_key_completions_(self):
        """Returns a tuple of keys.
        Special Method for ipython to get key completion
        """
        return tuple(self.keys())


def load(hdf, lazy=True, unpacker=_unpack_dataset, *args, **kwargs):
    """Returns a dictionary containing the
    groups as keys and the datasets as values
    from given hdf file.

    Parameters
    ----------
    hdf : string (path to file) or `h5py.File()` or `h5py.Group()`
    lazy : bool
        If True, the datasets are lazy loaded at the moment an item is requested.
    upacker : callable
        Unpack function gets `value` of type h5py.Dataset.
        Must return the data you would like to have it in the returned dict.

    Returns
    -------
    d : dict
        The dictionary containing all groupnames as keys and
        datasets as values.
    """

    def _recurse(hdfobject, datadict):
        for key, value in hdfobject.items():
            if type(value) == h5py.Group or isinstance(value, LazyHdfDict):
                if lazy:
                    datadict[key] = LazyHdfDict()
                else:
                    datadict[key] = {}
                datadict[key] = _recurse(value, datadict[key])
            elif isinstance(value, h5py.Dataset):
                if not lazy:
                    value = unpacker(value)
                datadict[key] = value

        return datadict

    with _hdf_file(hdf, lazy=lazy, *args, **kwargs) as hdf:
        if lazy:
            data = LazyHdfDict(_h5file=hdf)
        else:
            data = {}
        return _recurse(hdf, data)


def dump(data, hdf, packer=_pack_dataset, *args, **kwargs):
    """Adds keys of given dict as groups and values as datasets
    to the given hdf-file (by string or object) or group object.

    Parameters
    ----------
    data : dict
        The dictionary containing only string keys and
        data values or dicts again.
    hdf : string (path to file) or `h5py.File()` or `h5py.Group()`
    packer : callable
        Callable gets `hdfobject, key, value` as input.
        `hdfobject` is considered to be either a h5py.File or a h5py.Group.
        `key` is the name of the dataset.
        `value` is the dataset to be packed and accepted by h5py.

    Returns
    -------
    hdf : obj
        `h5py.Group()` or `h5py.File()` instance

    """

    def _recurse(datadict, hdfobject):
        for key, value in datadict.items():
            if isinstance(key, tuple):
                key = '_'.join((str(i) for i in key))
            if isinstance(value, (dict, LazyHdfDict)):
                hdfgroup = hdfobject.create_group(key)
                _recurse(value, hdfgroup)
            else:
                packer(hdfobject, key, value)

    with _hdf_file(hdf, *args, **kwargs) as hdf:
        _recurse(data, hdf)
        return hdf



"""Functionalities related to the navigation of hdf5 type files.
"""

import numpy as np
import h5py

from h5cross.utils import get_namedtuple_from_dict


def get_h5_structure(h5_filename, as_dict=True):
    """ Show hdf5 file components

        Input:
            :h5_filename: path to hdf5 file to inspect
    """
    # TODO: lazy?

    with h5py.File(h5_filename, "r") as node:
        out = log_hdf_node(node, as_dict=as_dict)

    return out


def log_hdf_node(node, as_dict=True):
    """
    Build a dictionary with the structure of a HDF5 node

    Parameters:
    -----------
    node : hdf5 node

    Returns:
    --------
    a dictionary
    """

    out = dict()

    def extend_dict(dict_, address, attr):
        tmp = dict_
        #!! Needs to be adapted for a multi-level nested object!!
        # Single level works fine, because we currently add attr to each level
        #  of the address, should be on the last level only -> 2 loops?
        for key in address[:]:
            if key not in tmp:
                #tmp[key] = dict()
                #print("Items current key and last key:", key, address[-1])
                if key == address[-1]:
                    #print("Last key", tmp)
                    if not as_dict:
                        attr_ = get_namedtuple_from_dict(attr, name='Info')
                    else:
                        attr_ = attr.copy()

                    tmp[key] = attr_
                else:
                    tmp[key] = {}
            tmp = tmp[key]  # so we redefine the dict to continue through

    def visitor_func(name, node):
        key_list = [item.strip() for item in name.split('/')]
        # print(key_list)
        if isinstance(node, h5py.Dataset):
            #print("Create dict with dtype and value")
            attr = dict()
            attr["dtype"] = str(node.dtype)
            attr["value"] = _get_node_description(node)
            extend_dict(out, key_list, attr)

        else:
            pass

    node.visititems(visitor_func)

    return out


def _ascii2string(ascii_list):
    """Ascii to string conversion.

    Parameters:
    -----------
    ascii_list : a list of string to be converted

    Returns:
    --------
    a string joining the list elements

    """
    return ''.join(chr(i) for i in ascii_list[:-1])


def _get_node_description(node):
    """Get number of elements in an array or value of a single-valued node.

    Parameters:
    -----------
    node : hdf5 node

    Returns:
    --------
    a value with a Python format
    None if data is not a singlevalued quantity
    """

    out = None
    value = node[()]
    shape = node.shape

    if np.prod(shape) == 1:
        # this is a strong assumption because if you find int8
        # your are probably looking at an hdf5 file applying the cgns standard
        # print(node)
        if node.dtype in ["int8"]:
            # TODO: rethink!
            out = np.char.array(_ascii2string(value))[0]
        elif shape in [(1), (1,)]:
            if node.dtype in ["int32", "int64"]:
                out = int(value[0])
            elif node.dtype in ["float32", "float64"]:
                out = float(value[0])
    else:
        out = "array of %s elements" % (" x ".join([str(k) for k in shape]))

    return out




# TODO: can be simplified by using meshio/yamio?
# TODO: move nek5000 to yamio?

import importlib

EXT_MODULE_MAP = {
    'vtk': 'vtk',
    'vtu': 'vtk',
    'pvtu': 'vtk',
    'nek5000': 'nek5000'

}


def _get_reader(file_format):
    if file_format not in EXT_MODULE_MAP:
        return None

    module_name = EXT_MODULE_MAP[file_format]
    module = importlib.import_module(f'h5cross.convert._{module_name}')

    return module.READER_MAP[file_format]


def convert(filename, file_format, **kwargs):
    convert_ = _get_reader(file_format)

    if convert_ is None:
        raise Exception("Format not available.")

    return convert_(filename, **kwargs)




import numpy as np
import seaborn as sns
import matplotlib.pyplot as plt

from h5cross.plotting._utils import _get_density
from h5cross.plotting._utils import _sort_by_density
from h5cross.plotting._utils import _get_scatter_title
from h5cross.plotting._utils import _get_bin_edges


def _plot_scatter(array1, array2, flag_xy_line=False):
    """Function to generate a scatter plot of two data arrays of same length with seaborn

    Input:
        array1: numpy array of first data set
        array2: numpy array of second data set
        axes_labels: list of string of length 2, used to specify the labels of output axis
        Optional Arguments:
            flag_save_: boolean controlling if the plots get saved (default = False)
            save_name_: (string), base name to be used in saving the plots (default = None)

    Output:
        Default = None, interactive view
        If flag_save = True then outputs a png image
    """

    sns.set_theme()

    density_array = _get_density(array1, array2)
    array1, array2, density_array = _sort_by_density(array1, array2,
                                                     density_array)

    data_plot = {"field1": array1, "field2": array2}
    joint_grid = sns.jointplot(data=data_plot, x="field1", y="field2",
                               kind="scatter", marginal_ticks=True,
                               marginal_kws=dict(stat="density"))

    joint_grid.plot_joint(plt.scatter, c=density_array, cmap="viridis")

    # draw line
    if flag_xy_line:
        xmin = np.min(np.append(array1, array2))
        xmax = np.max(np.append(array1, array2))
        joint_grid.ax_joint.plot([xmin, xmax], [xmin, xmax], "k-", linewidth=2)

    textstr = _get_scatter_title(array1, array2)
    joint_grid.fig.suptitle(textstr)

    return joint_grid.ax_joint


def _plot_distributions(array1, array2):
    """Function to generate histograms of two data arrays of different length with seaborn

    Input:
        array1: numpy array of first data set
        array2: numpy array of second data set
        axes_labels: list of string of length 2, used to specify the labels of output axis
        Optional Arguments:
            flag_save_: boolean controlling if the plots get saved (default = False)
            save_name_: (string), base name to be used in saving the plots (default = None)
            flag_show_: boolean setting if the plot gets showed or not (default = True)

    Output:
        Default = None, interactive view
        If flag_save = True then outputs a png image
    """

    sns.set_theme()
    # Will be deprecated soon, must update data structure
    # https://seaborn.pydata.org/tutorial/data_structure.html
    data_plot = {"field1": array1, "field2": array2}

    bin_edges = _get_bin_edges(array1, array2)

    ax = sns.histplot(data_plot, bins=bin_edges,
                      palette=["cornflowerblue", "lightgreen"])

    return ax
